﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2017 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System.Threading;
using System.Threading.Tasks;
using HeuristicLab.Collections;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Core.Networks;
using HeuristicLab.Data;
using HeuristicLab.Optimization;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Networks.IntegratedOptimization {
  [Item("OrchestratorNode", "An orchestrator node.")]
  [StorableClass]
  public abstract class OrchestratorNode : Node, IOrchestratorNode {
    [Storable]
    private ParameterCollection parameters;
    protected ParameterCollection Parameters {
      get { return parameters; }
    }
    private ReadOnlyKeyedItemCollection<string, IParameter> readOnlyParameters;
    IKeyedItemCollection<string, IParameter> IOrchestratorNode.Parameters {
      get {
        if (readOnlyParameters == null) readOnlyParameters = parameters.AsReadOnly();
        return readOnlyParameters;
      }
    }

    [Storable]
    private ResultCollection results;
    public ResultCollection Results {
      get { return results; }
    }

    [StorableConstructor]
    protected OrchestratorNode(bool deserializing) : base(deserializing) { }
    protected OrchestratorNode(OrchestratorNode original, Cloner cloner) : base(original, cloner) {
      results = cloner.Clone(original.results);
      parameters = cloner.Clone(original.parameters);
      readOnlyParameters = null;
      RegisterEvents();
    }

    protected OrchestratorNode() : base("OrchestratorNode") {
      results = new ResultCollection();
      parameters = new ParameterCollection();
      readOnlyParameters = null;
    }
    protected OrchestratorNode(string name) : base(name) {
      results = new ResultCollection();
      parameters = new ParameterCollection();
      readOnlyParameters = null;
    }
    protected OrchestratorNode(string name, string description) : base(name, description) {
      results = new ResultCollection();
      parameters = new ParameterCollection();
      readOnlyParameters = null;
    }

    [StorableHook(HookType.AfterDeserialization)]
    private void AfterDeserialization() {
      RegisterEvents();
    }

    private void RegisterEvents() {
      foreach (var p in Ports) RegisterPortEvents(p);
    }

    protected IMessagePort CreateOrchestrationPort<T>(string portName) where T : class, IProblem {
      var orchestrationPort = new MessagePort(portName);
      orchestrationPort.Parameters.Add(new PortParameter<EnumValue<OrchestrationMessage>>("OrchestrationMessage") {
        Type = PortParameterType.Output
      });
      orchestrationPort.Parameters.Add(new PortParameter<T>("Problem") {
        Type = PortParameterType.Output
      });
      orchestrationPort.Parameters.Add(new PortParameter<ResultCollection>("Results") {
        Type = PortParameterType.Input
      });
      return orchestrationPort;
    }

    protected IMessagePort CreateEvaluationPort<T>(string portName, string solutionName, string qualityName) where T : class, IItem {
      var evaluationPort = new MessagePort(portName);
      evaluationPort.Parameters.Add(new PortParameter<T>(solutionName) {
        Type = PortParameterType.Input
      });
      evaluationPort.Parameters.Add(new PortParameter<DoubleValue>(qualityName) {
        Type = PortParameterType.Input | PortParameterType.Output
      });
      return evaluationPort;
    }

    protected virtual void ProcessMessage(IMessage message, IMessagePort port, CancellationToken token) { }

    public abstract void Prepare(bool clearRuns = false);
    public async Task PrepareAsync(bool clearRuns = false) { await Task.Run(() => Prepare(clearRuns)); }
    public abstract void Start();
    public async Task StartAsync() { await Task.Run(() => Start()); }
    public abstract void Pause();
    public async Task PauseAsync() { await Task.Run(() => Pause()); }
    public abstract void Stop();
    public async Task StopAsync() { await Task.Run(() => Stop()); }

    #region Events
    protected override void Ports_ItemsAdded(object sender, CollectionItemsChangedEventArgs<IPort> e) {
      base.Ports_ItemsAdded(sender, e);
      foreach (var p in e.Items)
        RegisterPortEvents(p);
    }
    protected override void Ports_ItemsRemoved(object sender, CollectionItemsChangedEventArgs<IPort> e) {
      base.Ports_ItemsRemoved(sender, e);
      foreach (var p in e.Items)
        DeregisterPortEvents(p);
    }
    protected override void Ports_ItemsReplaced(object sender, CollectionItemsChangedEventArgs<IPort> e) {
      base.Ports_ItemsReplaced(sender, e);
      foreach (var p in e.OldItems)
        DeregisterPortEvents(p);
      foreach (var p in e.Items)
        RegisterPortEvents(p);
    }
    protected override void Ports_CollectionReset(object sender, CollectionItemsChangedEventArgs<IPort> e) {
      base.Ports_CollectionReset(sender, e);
      foreach (var p in e.OldItems)
        DeregisterPortEvents(p);
      foreach (var p in e.Items)
        RegisterPortEvents(p);
    }
    private void MessagePort_MessageReceived(object sender, EventArgs<IMessage, CancellationToken> e) {
      ProcessMessage(e.Value, (IMessagePort)sender, e.Value2);
    }
    #endregion

    #region Port Events
    protected virtual void RegisterPortEvents(IPort port) {
      var mp = port as IMessagePort;
      if (mp != null)
        mp.MessageReceived += MessagePort_MessageReceived;
    }
    protected virtual void DeregisterPortEvents(IPort port) {
      var mp = port as IMessagePort;
      if (mp != null)
        mp.MessageReceived -= MessagePort_MessageReceived;
    }
    #endregion
  }
}
